@@ -1,7 +1,7 @@
 from datetime import datetime
 from typing import List as TypedList
 
-from django.db import models
+from django.db import models, transaction
 from django.db.models.functions import Length
 
 from userdefinedtables.settings import USER_DEFINED_TABLES_DECIMAL_FIELD_KWARGS
@@ -56,35 +56,50 @@ class Row(models.Model):
         null=False,
         blank=False,
     )
-    previous_row = models.OneToOneField(
-        "userdefinedtables.row",
-        on_delete=models.CASCADE,
-        related_name="_nxt_row",
-        null=True,
-    )
-    next_row = models.OneToOneField(
-        "userdefinedtables.row",
-        on_delete=models.CASCADE,
-        related_name="_prv_row",
-        null=True,
-    )
+    index = models.PositiveBigIntegerField()
 
     class Meta:
         constraints = [
             models.UniqueConstraint(
-                fields=["list", "previous_row"],
-                name="For any list, a row can only follow one or zero rows. "
-                "A null row can be followed by exactly one row, per list.",
-                condition=models.Q(previous_row__isnull=True),
-            ),
-            models.UniqueConstraint(
-                fields=["list", "next_row"],
-                name="For any list, a row can only precede one or zero rows. "
-                "A null row can be preceded by exactly one row, per list.",
-                condition=models.Q(next_row__isnull=True),
+                fields=["list", "index"],
+                name="A list can only have a single row at an index.",
+                deferrable=models.Deferrable.DEFERRED,
             ),
         ]
 
+    def insert_at(self, index):
+        assert not self._state.adding, "Insertions on new items are done implicitly. This method is the approach to use"
+        " when an item already exists and is meant to be inserted in another location."
+
+        # in the trivial case, we do not actually need to move things around, so we return
+        if self.index == index:
+            return
+
+        with transaction.atomic():
+            # in either of the other cases, we need to move things around, and so wrap either move in a transaction
+            if self.index < index:
+                Row.objects.filter(index__gt=self.index, index__lte=index).update(index=models.F("index") - 1)
+            else:
+                Row.objects.filter(index__lt=self.index, index__gte=index).update(index=models.F("index") + 1)
+            self.index = index
+            self.save()
+
+    def save(self):
+        if self._state.adding:
+            # if there are no rows indexed above this one, this will perform no changes, but will still make a db query
+            Row.objects.filter(list=self.list, index__gte=self.index).update(index=models.F("index") + 1)
+        # in a situation where we are not adding a row, we will need to handle the change directly using
+        # `self.insert_at`
+        super().save()
+
+    def delete(self):
+        index = self.index
+        list = self.list
+        super().delete()
+        higher_rows = Row.objects.filter(list=list, index__gte=index)
+        if higher_rows.exists():
+            higher_rows.update(index=models.F("index") - 1)
+
 
 class Entry(models.Model):
     row = models.ForeignKey(
